/*
 ******************************************************************************
 ******************************************************************************
 *****                             PEXtk                                  *****
 *****									  *****
 *****		    An Immediate Mode Interface to PEX			  *****
 *****									  *****
 *****			    Pick Example Program    			  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1993, All Rights Reserved                                   *
 *			SHOgraphics					      *
 *			321 Soquel Way 			      		      *
 *			Sunnyvale, CA  94086       		              *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of ShoGraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED LIMITED TO SPECIAL, INDIRECT OR 
 CONSEQUENTIAL DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE
 PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS 
 ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 SOFTWARE.

 ************************************************************************ 
 *		PEXtk Example Program for picking
 *
 *	This program is an example for using picking. It will draw a 'car' 
 *	using a circle and 'spokes' to simulate the wheels.
 *	It will then move a 'pick box' across the screen demonstrate picking.
 *
 *	Each 'wheel' will have its 'parent' pick id shifted and ored with the 
 *	wheel number.
 *
 */
#include <stdio.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <PEXtk.h>

#define NUM_SPOKES	4
#define NUM_WHEELS	4
#define BUF_SIZE	100
    int			incr, ViewType;
    int			pause_flag;
    int			vct, poly_cnt;
    float		radius;
    float		near, far;
    float		left, right, bottom, top, fov;
    GC			gc;
    unsigned long	white_pix;
    int 		win_width=600, win_height=600;

Fcoord3D verts[2*NUM_SPOKES];
Fcoord3D loc[NUM_WHEELS];

main(argc, argv)
int argc; char **argv;
{
    int i,j,k,m,n, win_width=600, win_height=600;
    int x,y,xsize,ysize,max_hits,moreRet;
    int buffer[BUF_SIZE];
    Display 		 *dpy;
    Window 		 window;
    int		 	 *infoData;

    ViewType = 0; pause_flag = 0;
    incr = 100;
    left = -5.0; right = 5.0; bottom = -5.0; top = 5.0;
    near = 1.0; far = 1000.0; fov = 90.0;
    radius = 0.8;

    parse_line(argc,argv);

    if ((dpy = XOpenDisplay ("")) == NULL) {
	printf(" Error opening display\n");
	return 1;
    }

    if(Pinit(dpy)){
	printf("ERROR... initializing PEX server\n");
	exit(1);
    }

    if( (window = Pcreate_window(dpy, win_width, win_height)) == NULL){
	printf("ERROR... creating window\n");
	exit(1);
    }

    init_data();
    vct = NUM_SPOKES;
    gc = DefaultGC(dpy,DefaultScreen(dpy));
    white_pix = WhitePixel(dpy,DefaultScreen(dpy));

    Pset_drawable( window );
    Pzbuffer(1);

	/* setup defaults */
	Pbuffer_mode(  PEXtk_DOUBLE_BUFFER );
	Pviewport( 0, win_width-1, 0, win_height-1);
	Pbackface( PEXtk_CULL_BACKFACE);
	Ptext_size( .15, .15);
	Pcolor_rgb( 0.9, 0.0, 0.4);
	Pline_color_rgb( 0.9, 0.0, 0.4);
	Pmarker_type( PEXtk_MRKR_X);
	Pmarker_color_rgb( 0.0, 1.0, 0.0);

    Set_view();

    xsize = ysize = 20;
    max_hits = 10000;
    x = xsize/2; y = win_height / 2 - 80;

    while( x < win_width ){
	Pclear_rgb( 0.0, 0.0, 0.0);

	/* "Draw" the objects now to pick them */
        Pbegin_pick( x, y, xsize, ysize, max_hits );
	  draw_car();
        n = Pend_pick( buffer, BUF_SIZE, &moreRet);

	/* Draw the objects now to put on screen */
        draw_car();
	/* Draw the 'pick box' */
	XSetForeground( dpy, gc, white_pix);
	XDrawRectangle(dpy, Pget_backbuffer(), gc, x-xsize/2,
        	win_height - y - ysize/2, xsize, ysize);
        Pswap_buffers();

	/* If there was any hits, then print the pick buffer contents */
	if( n != 0 ){
	    printf(" Got %d hits\n",n);
	    j = 0;
	    while(n > 0){
		k = buffer[j]; j++;
		for(m=0;m<k;m++,j++)
		    printf(" %d, %d ",buffer[j]>>16,buffer[j]&0xffff);
		printf("\n");
		n--;
	    }
	    sleep(3);
	}
        x = x + incr; /* move pick box across screen */
    } /* end while */

    if( pause_flag ){
	printf("enter a char ");
	fscanf( stdin, "%c",&i);
    }
    Pexit();
    exit(0);
}

draw_spoke(i)
int i;
{
    Ppoly_line( 2, &verts[2*i].x);
}

draw_wheel(id, n)
int id,n;
{
    int i,j;
    Ppick_id( id );
    Pcircle( 0., 0., -5.0, radius);
    for(i=0;i<n;i++){
	Ppick_id( id << 16 | (i+1));
	draw_spoke(i);
    }
}

draw_car()
{
    int i;
    Pload_identity();
    for(i=0;i<NUM_WHEELS;i++){
	Ppush_matrix();
	Ptranslate( loc[i].x, loc[i].y, loc[i].z, 1);
	draw_wheel(i+1, NUM_SPOKES);
	Ppop_matrix();
    }
}

Set_view()
{
    int i, j;
    switch( ViewType ){
	case 0:
	    Ppersp_view( 0, fov, 1.0, near, far);
	    break;
	case 1:
	    Portho_view( 0, left, right, bottom, top, near, far);
	    break;
	case 2:
	    Portho_2d_view( 0, left, right, bottom, top );
	    break;
	case 3:
	    Pwindow_view( 0, left, right, bottom, top, near, far );
	    break;
	case 4:
	    i = right - left;
	    j = top - bottom;
	    printf("viewport  0.. %d, 0.. %d\n",i,j);
	    Pviewport( 0, (int)(i), 0, (int)(j));
	    Pwindow_view( 0, left, right, bottom, top, near, far );
	    break;
    }
}

init_data()
{
    verts[0].x =  0.0; verts[0].y = 0.0;     verts[0].z = -5.0;
    verts[1].x =  radius; verts[1].y = 0.0;  verts[1].z = -5.0;
    verts[2].x =  0.0; verts[2].y = 0.0;     verts[2].z = -5.0;
    verts[3].x = 0.0;  verts[3].y = radius;  verts[3].z = -5.0;
    verts[4].x =  0.0; verts[4].y = 0.0;     verts[4].z = -5.0;
    verts[5].x = -radius; verts[5].y = 0.0;  verts[5].z = -5.0;
    verts[6].x =  0.0; verts[6].y = 0.0;     verts[6].z = -5.0;
    verts[7].x = 0.0;  verts[7].y = -radius; verts[7].z = -5.0;

    loc[0].x =  4.0*radius; loc[0].y = 2.0*radius;  loc[0].z = 0.0;
    loc[1].x = -4.0*radius; loc[1].y = 2.0*radius;  loc[1].z = 0.0;
    loc[2].x = -4.0*radius; loc[2].y = -2.0*radius; loc[2].z = 0.0;
    loc[3].x = 4.0*radius;  loc[3].y = -2.0*radius; loc[3].z = 0.0;
}

/*
 *                      +---------------------------+
 * ---------------------+         parse_line        +------------------------
 *                      +---------------------------+
 *  FUNCTION:
 *      Parses the command line.
 *
 *  USAGE:
 *      parse_line(argc, argv)
 *	int	   argc 	   (in)  - number of arg's
 *	char   	   argv 	   (in)  - pointer to arg list
 *
 *  DESCRIPTION:
 *	This routine will parse the command line for the following options:
 *	   opt  argument	type	  comment
 *	   ---  ------------    ----      ----------------------------------
 *	    -i incr		int	  increment (in pixels) to move pick box
 *	    -m mode		int	  mode: 0=persp,1=ortho,2=ortho2,
 *					  3=window
 *	    -p 			void	  sets pause flag
 *	    -w width		float	  Sets window size
 *	    -z trans		float	  z translation value
 *	    -h near		float	  near (hither) plane
 *	    -y far		float	  far (yon) plane
 *	    -l left		float	  left
 *	    -r right		float	  right
 *	    -t top		float	  top
 *	    -b bottom		float	  bottom
 *                                
*/
int parse_line(argc,argv)
int argc;
char **argv;
{
     char *temp,ffile[32];
     extern char *optarg;
     extern int optind;
     int c;

    while((c=getopt(argc,argv,"a:i:h:y:l:r:b:t:m:t:pvw:")) != EOF)
    {
	switch(c){
	    case 'i':
		  sscanf(optarg,"%d",&incr);
		break;
	    case 'm':
		  sscanf(optarg,"%d",&ViewType);
		break;
	    case 'p':
		  pause_flag = 1;
		break;
	    case 'w':
		  sscanf(optarg,"%f",&win_width);
		  if( win_width > 0 && win_width < 1024)
			win_height = win_width;
		  else 
			win_width = win_height;
		break;
	    case 'h':
		  sscanf(optarg,"%f",&near);
		break;
	    case 'y':
		  sscanf(optarg,"%f",&far);
		break;
	    case 'l':
		  sscanf(optarg,"%f",&left);
		break;
	    case 'r':
		  sscanf(optarg,"%f",&right);
		break;
	    case 'b':
		  sscanf(optarg,"%f",&bottom);
		break;
	    case 't':
		  sscanf(optarg,"%f",&top);
		break;
	}
    }

}

Usage()
{
    printf("Error...  Usage is:\n");
    printf(" pck_test [options]\n");
    printf(" options: \n");
    printf("\t -m mode {0-3}\t(view type,0=persp,1=ortho)\n");
    printf("\t\t\t(2=ortho2,3=window)\n");
    printf("\t -i incr\t(number of pixels to move pick box by)\n");
    printf("\t -r radius\t(right coord, float)\n");
    printf("\t -x value\t(x start value pick box)\n");
    printf("\t -y value\t(y start value pick box)\n");
    printf("\t -w width\t(sets width,height of window)\n");
    printf("\t -p\t\t(pause flag)\n");
}
