/*
 *  $Header: /auto/nfs/sequoia/u/an/work/rws/pextk/src/pextk_utl.c,v 1.3 1994/01/06 00:27:05 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                          pextk_utl.c                               *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk utility primitive
 *	building routines.
 *
 *  Module Notes:
 *
 *  Module Contents:
 *
 *    int   PEXtk_init_utl    Initializes the required pointers
 *    void  Parc              Creates an arc by using polylines.
 *    void  Parc_fill         Creates a filled arc.
 *    void  Pcircle           Creates a circle outline by using polylines.
 *    void  Pcircle_fill      Creates a filled circle.
 *    void  Psphere           Creates a sphere by using polylines.
 *    void  Psphere_fill      Creates a filled sphere.
 *
 *    int   PEXtk_calc_vertex_normals	Calculates vertex normals given
 *					list of verts's, polygon list.
 *    int   PEXtk_calc_polygon_normals
 *
 *  Module Description:
 *
 */

#undef FUNCPROTO

#include <stdio.h>
#include <math.h>

#ifndef M_PI
#define M_PI 3.14159265359
#endif

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>

#ifdef MPEX
#include <X11/MPEX.h>
#include <X11/MPEXproto.h>
#include <X11/MPEXlib.h>
#endif

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_lcl.h"

	int PEXtk_build_arc( );
	int PEXtk_build_circle( );

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_init_utl      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes utility primitive building pointers.
 *
 *  USAGE:
 *	PEXtk_init_utl( p_obj )
 *	PEXtk_OBJECT    *p_obj	(in)	Pointer to PEXtk object.
 *
 *  DESCRIPTION:
 *	This routine initializes pointers for p_build and p_state.
 *
 *  RETURNS:
 *	0 if successful, 1 if error
 */
 
/*ARGSUSED*/
int
PEXtk_init_utl(p_obj)
PEXTK_OBJECT		*p_obj;
{
    int i,j;
    /*
     * Save the various object ptrs
     */
    p__object	= p_obj;
    p_state	= p_obj->state;
    p_build	= p_obj->build;
    p_buffer	= p_obj->buffer;
    p_support	= p_obj->support;
    p_dl	= p_obj->dpy_list;
    p_clr	= p_obj->clr_tbl;
    p_x_data	= p_obj->x_data;

    dpy = p_x_data->dpy;

	/* get memory for the arc and circle indicies buffer */
	i = (int) p_state->attr->u_prec;
	j = (int) p_state->attr->v_prec;
	PEXtk_build_cosine_buffer(i, j);

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_close_utl     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Closes utilities routines (frees data).
 *
 *  USAGE:
 *	PEXtk_close_utl( )
 *
 *  DESCRIPTION:
 *	This routine de-allocates memory for the items allocated in
 *	the PEXtk_init_utl routine.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_close_utl()
{

    if( p_buffer->indicies != NULL )
	free( p_buffer->indicies);

    if( p_build->cos_buffer_u != NULL )
	free( p_build->cos_buffer_u);
    if( p_build->sin_buffer_u != NULL )
	free( p_build->sin_buffer_u);

    if( p_build->cos_buffer_v != NULL )
	free( p_build->cos_buffer_v);
    if( p_build->sin_buffer_v != NULL )
	free( p_build->sin_buffer_v);

}
    
/*
 *
 *                       +--------------------------+
 * ----------------------+            Parc          +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Creates an arc by using polylines.
 *
 *  USAGE:
 *	Parc( x, y, z, radius, start, end )
 *	Fcoord	    x		(in)	The x position for the arc
 *	Fcoord	    y		(in)	The y position for the arc
 *	Fcoord	    z		(in)	The z position for the arc
 *	Fcoord	    radius	(in)	The radius for the arc
 *	Angle	    start	(in)	The starting angle for the arc
 *	Angle	    end		(in)	The ending angle for the arc
 *
 *  NOTE:
 *	The start and end angles are referenced from the x-axis.
 *	The arc is polylines from the start angle to the end angle,
 *	and no connection is made from the end point to the center
 *	or to the start point.
 *
 *	This routine simply calls the build routine PEXtk_build_arc.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Parc( x, y, z, radius, start, end )
Fcoord	x;
Fcoord	y;
Fcoord	z;
Fcoord	radius;
Angle	start;
Angle	end;
{

	PEXtk_build_arc( 0, x, y, z, radius, start, end);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Parc_fill       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Creates a filled arc.
 *
 *  USAGE:
 *	Parc_fill( x, y, z, radius, start, end )
 *	Fcoord	    x		(in)	The x position for the arc
 *	Fcoord	    y		(in)	The y position for the arc
 *	Fcoord	    z		(in)	The z position for the arc
 *	Fcoord	    radius	(in)	The radius for the arc
 *	Angle	    start	(in)	The starting angle for the arc
 *	Angle	    end		(in)	The ending angle for the arc
 *
 *  NOTE:
 *	The start and end angles are referenced from the x-axis.
 *	The arc_fill is a polygon from the start angle to the end angle,
 *	and a connection is made from the start point to the center
 *	and the end point to the center.
 *
 *	This routine simply calls the build routine PEXtk_build_arc.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Parc_fill( x, y, z, radius, start, end )
Fcoord	x;
Fcoord	y;
Fcoord	z;
Fcoord	radius;
Angle	start;
Angle	end;
{
	PEXtk_build_arc( 1, x, y, z, radius, start, end);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_build_arc     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Builds the arc primitive
 *
 *  USAGE:
 *	PEXtk_build_arc( type, x, y, radius, start, end )
 *	int	    type	(in)	The type of primitive to use,
 *					    0=polyline, 1=polygon.
 *	Fcoord	    x		(in)	The x position for the arc
 *	Fcoord	    y		(in)	The y position for the arc
 *	Fcoord	    z		(in)	The z position for the arc
 *	Fcoord	    radius	(in)	The radius for the arc
 *	Angle	    start	(in)	The starting angle for the arc
 *	Angle	    end		(in)	The ending angle for the arc
 *
 *  NOTE:
 *	No attempt has been made to optimize this routine.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_build_arc( type, x, y, z, radius, start, end )
int	type;
Fcoord	x;
Fcoord	y;
Fcoord	z;
Fcoord	radius;
Angle	start;
Angle	end;
{

    int 	i, res;
    float 	inc;
    PEXCoord 	*buf, *ptr, ptr2;
    double 	rad_start, rad_end, cur_deg, delta_deg;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	res = (int) p_state->attr->u_prec;
	rad_start = start * M_PI / 180.0;
	rad_end   = end   * M_PI / 180.0;

	i = (res+3)*sizeof(PEXCoord);
	if(DB_check_build_buffer_size(i)) return 1;
	buf = (PEXCoord *) p_build->buffer;

	ptr = buf;
	if( type == 1 ){ /* for an arc_fill, put in center point */
	    ptr->x = x; ptr->y = y; ptr->z = z;
	    ptr++;
	}
	delta_deg = (rad_end - rad_start ) / (double) res;
	for( i = 0, cur_deg = rad_start; i < res; i++, ptr++){
	    ptr->x = x + radius*cos( cur_deg);
	    ptr->y = y + radius*sin( cur_deg);
	    ptr->z = z;
	    cur_deg += delta_deg;
	}
	/*
	 * Do last point separate, to avoid any accumulated error
	 */
	    ptr->x = x + radius*cos( rad_end);
	    ptr->y = y + radius*sin( rad_end);
	    ptr->z = z;
	    ptr++;

	ptr = buf;
	if( type == 0 ){
	    PEXPolyline( dpy, p_state->cur_id, p_state->req_type, res+1, buf);
	} else {
	    ptr->x = x; ptr->y = y; ptr->z = z;
	    Ppoly_fill( res+2, (Fcoord3D *) buf );
	}
}

/*
 *
 *                       +--------------------------+
 * ----------------------+           Pcircle        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Creates a circle outline by using polylines.
 *
 *  USAGE:
 *	Pcircle( x, y, z, radius )
 *	Fcoord	    x		(in)	The x position for the circle
 *	Fcoord	    y		(in)	The y position for the circle
 *	Fcoord	    z		(in)	The z position for the circle
 *	Angle	    radius	(in)	The radius for the circle
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pcircle( x, y, z, radius )
Fcoord	x;
Fcoord	y;
Fcoord	z;
Angle	radius;
{
	PEXtk_build_circle( 0, x, y, z, radius );
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pcircle_fill      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Creates a filled circle.
 *
 *  USAGE:
 *	Pcircle_fill( x, y, z, radius )
 *	Fcoord	    x		(in)	The x position for the circle
 *	Fcoord	    y		(in)	The y position for the circle
 *	Fcoord	    z		(in)	The z position for the circle
 *	Angle	    radius	(in)	The radius for the circle
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pcircle_fill( x, y, z, radius )
Fcoord	x;
Fcoord	y;
Fcoord	z;
Angle	radius;
{
	PEXtk_build_circle( 1, x, y, z, radius );
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     PEXtk_build_circle   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Builds the circle primitive
 *
 *  USAGE:
 *	PEXtk_build_circle( type, x, y, radius )
 *	int	    type	(in)	The type of primitive to use,
 *					    0=polyline, 1=polygon.
 *	Fcoord	    x		(in)	The x position for the arc
 *	Fcoord	    y		(in)	The y position for the arc
 *	Fcoord	    z		(in)	The z position for the arc
 *	Fcoord	    radius	(in)	The radius for the arc
 *
 *  NOTE:
 *	No attempt has been made to optimize this routine.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_build_circle( type, x, y, z, radius )
int	type;
Fcoord	x;
Fcoord	y;
Fcoord	z;
Fcoord	radius;
{

    int 	i, res;
    float 	inc;
    PEXCoord 	*buf, *ptr;
    double 	cur_deg, delta_deg;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	res = (int) p_state->attr->u_prec;
	PEXtk_build_cosine_buffer(res, 0);

	i = (res+1)*sizeof(PEXCoord);
	if(DB_check_build_buffer_size(i)) return 0;
	buf = (PEXCoord *) p_build->buffer;

	ptr = buf;
	delta_deg = (double) ((double) 2.0 * M_PI  / (double) res);

	for( i = 0, cur_deg = 0.0; i < res; i++, ptr++){
	    ptr->x = x + radius*p_build->cos_buffer_u[i];
	    ptr->y = y + radius*p_build->sin_buffer_u[i];
	    cur_deg += delta_deg;
	    ptr->z = z;
	}

	if( type == 0 ){
	    ptr->x = buf->x; ptr->y = buf->y; ptr->z = z;
	    PEXPolyline( dpy, p_state->cur_id, p_state->req_type, res+1, buf);
	} else
	    Ppoly_fill( res, (Fcoord3D *) buf );

}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Psphere         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Creates an sphere by using polylines.
 *
 *  USAGE:
 *	Psphere( x, y, z, radius )
 *	Fcoord	    x		(in)	The x position for the sphere
 *	Fcoord	    y		(in)	The y position for the sphere
 *	Fcoord	    z		(in)	The z position for the sphere
 *	Fcoord	    radius	(in)	The radius for the sphere
 *
 *  NOTE:
 *	No attempt has been made to optimize this routine.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Psphere( x, y, z, radius )
Fcoord	x;
Fcoord	y;
Fcoord	z;
Fcoord	radius;
{
	
    int 	i, j, n, res_u, res_v, last_cnt=1;
    float 	f_const;
    PEXCoord 	*buf, *ptr, *f_ptr;
    double 	cur_deg_u, cur_deg_v, delta_deg_u, delta_deg_v, cur_rad;
    double 	sin_cur_u, cos_cur_u, sin_cur_v, cos_cur_v;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	res_u = (int)p_state->attr->u_prec;
	res_v = (int)p_state->attr->v_prec;
	PEXtk_build_cosine_buffer(res_u, res_v);

	/* calculate the amount of memory needed */
	/* all vertices are save */
	n = (res_u*res_v+2)*sizeof(PEXCoord);
	if(DB_check_build_buffer_size(n)) return;
	buf = (PEXCoord *) p_build->buffer;

	n = (res_v+1)*sizeof(PEXCoord);
	if(DB_check_bld_vert(n)) return;
	/*
	 * v runs from top pole to bottom pole, 180 degrees.
	 * u runs from angle = 0 to angle = 360.
	 */
	delta_deg_u = (double) ((double) 2.0 * M_PI  / (double) res_u);
	delta_deg_v = (double) ((double) M_PI  / (double) res_v);

	cur_deg_v = 0.0;
#ifdef GEN_LONG
	f_ptr = buf;
	ptr = buf;
	for( i = 0; i < res_u/2; i++){
	    cur_deg_v = 0.0;
	    sin_cur_u = sin(cur_deg_u);
	    cos_cur_u = cos(cur_deg_u);
	    for( j = 0; j < res_v; j++){
		ptr->x = x + radius*cos( cur_deg_v) * cos_cur_u;
	        ptr->y = y + radius*sin( cur_deg_v);
		ptr->z = z - radius*(cos( cur_deg_v) * sin_cur_u);
		cur_deg_v += delta_deg_v;
		ptr++;
	    }
	    ptr->x = f_ptr->x; ptr->y = f_ptr->y; ptr->z = f_ptr->z;
	    PEXPolyline( dpy, p_state->cur_id, p_state->req_type, res_u+1, f_ptr);
	    /* 
	     *	Use this position for next goup of lines.
	     *  Don't save the duplicated first point.
	     */
	    cur_deg_u += delta_deg_u;

	    f_ptr = ptr;
	}
#endif

	/*
	 * Do the horizontal/circumference lines.
	 * These are generated from the top pole position, past
	 * the 'equator', to the bottom pole position.
	 */
	cur_rad = 0.0;
	ptr = buf;
	/* calculate the top pole position */
	ptr->x = x; ptr->y = y + radius; ptr->z = z; ptr++;
	cur_deg_v = delta_deg_v;

	for( i = 0; i < res_v-1; i++){
	    n = (i+1) % res_v;
	    cos_cur_v = p_build->cos_buffer_v[i];
	    sin_cur_v = p_build->sin_buffer_v[i];
	    cur_rad = radius*sin_cur_v;
	    f_const = y + radius*cos_cur_v;
	    cur_deg_u = 0.0;
	    f_ptr = ptr;
	    for( j = 0; j < res_u; j++ ){
	        n = (j+1) % res_u;
		ptr->x = x + cur_rad*p_build->cos_buffer_u[n];
		ptr->y = f_const;
		ptr->z = z - cur_rad*p_build->sin_buffer_u[n];
		cur_deg_u += delta_deg_u;
		ptr++;
	    }
	    /* Don't save duplicated first point */
	    ptr->x = f_ptr->x; ptr->y = f_ptr->y; ptr->z = f_ptr->z;
	    PEXPolyline( dpy, p_state->cur_id, p_state->req_type, res_u+1, f_ptr);
	    cur_deg_v += delta_deg_v;
	}

	/*
	 * Build the 'longitudinal' lines by copying them from the
	 * previously calculated data.
	 */
	ptr = (PEXCoord *) p_build->bld_vert;
	ptr->x = x; ptr->y = y + radius; ptr->z = z;
	ptr = ptr + res_v;
	ptr->x = x; ptr->y = y - radius; ptr->z = z;

	for( i = 0; i < res_u; i++){
	    f_ptr = buf+i+1;
	    ptr = (PEXCoord *) p_build->bld_vert;
	    ptr++;
	    /* go from top pole to bottom pole */
	    for( j = 0; j < res_v-1; j++, ptr++ ){
		memcpy( ptr, f_ptr, sizeof(PEXCoord));
		f_ptr += res_u;
	    }
	    PEXPolyline( dpy, p_state->cur_id, p_state->req_type, 
		res_v+1, p_build->bld_vert);
	}
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Psphere_fill      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Creates a filled sphere.
 *
 *  USAGE:
 *	Psphere_fill( x, y, z, radius )
 *	Fcoord	    x		(in)	The x position for the sphere
 *	Fcoord	    y		(in)	The y position for the sphere
 *	Fcoord	    z		(in)	The z position for the sphere
 *	Fcoord	    radius	(in)	The radius for the sphere
 *
 *  RETURNS:
 *	Nothing
 *
 *  DESCRIPTION:
 *	This routine will build a sphere. Each of the 'poles' will be
 *	created using SOFAS, and the middle portion will be built using
 *	triangle strips.
 */
 
/*ARGSUSED*/
void
Psphere_fill( x, y, z, radius )
Fcoord	x;
Fcoord	y;
Fcoord	z;
Fcoord	radius;
{

    int 	i, j, n, res_u, res_v, last_cnt=1;
    float 	f_const;
    PEXCoord 	*buf, *ptr, *f_ptr;
    double 	cur_deg_u, cur_deg_v, delta_deg_u, delta_deg_v, cur_rad;
    double 	sin_cur_u, cos_cur_u, sin_cur_v, cos_cur_v;
    unsigned short 	*ndx;
    PEXArrayOfVertex 	vertList;
    PEXArrayOfFacetData facetList;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

#ifdef KPC
    if( p_support->gdp_sphere ){
	PEXCoord ctr;
	float rad;
	ctr.x = x; ctr.y = y; ctr.z = z;
	rad = radius;
	KPEXUnpackedSphereList( dpy, p_state->cur_id, p_state->req_type, 1,
		&ctr, &rad, NULL, NULL);
    } else {
#endif
	res_u = (int)p_state->attr->u_prec;
	res_v = (int)p_state->attr->v_prec;
	PEXtk_build_cosine_buffer(res_u, res_v);

	/* calculate the amount of memory needed */
	n = (res_u*2+2)*sizeof(PEXCoord);
	if(DB_check_build_buffer_size(n)) return;

	buf      = (PEXCoord *) p_build->buffer;
	vertList.no_data = (PEXCoord *) p_build->buffer;

	/*
	 * v runs from top pole to bottom pole, 180 degrees.
	 * u runs from angle = 0 to angle = 360.
	 */
	delta_deg_u = (double) ((double) 2.0 * M_PI  / (double) res_u);
	delta_deg_v = (double) ((double) M_PI  / (double) res_v);

	/* 
	 * Check the poly_Count and indicies buffers size.
	 * This is used for the indexed polys at the top pole
	 * and bottom pole. Triangles are used for this.
	 * The poly_Count buffer using indexed polys must be at least
	 * res_u in size. That is max number of polygons used.
	 * The indicies buffer must be at least res_u*3 in size.
	 */
	i = 2 * res_u;
	/* DB_check_poly_count_list( i); */
	DB_check_connectivity_size( i);
	i = 3 * res_u;

	DB_check_indicies_size( i);
	ndx = p_buffer->indicies;

	PEXtk_check_poly_attrib( PEXInteriorStyleSolid, 0);

	/*
	 * Do the horizontal/circumference lines.
	 * These are generated from the top pole position, past
	 * the 'equator', to the bottom pole position.
	 */
	cur_rad = 0.0;
	ptr = buf;

	/* calculate the top pole position */
	ptr->x = x; ptr->y = y + radius; ptr->z = z; ptr++;

	/*
	 * Do the top set of triangles.
	 */
	cur_deg_u = delta_deg_u;
	f_ptr = ptr;

	/* do the indicies for the top section */

	for(i=0,j=0; i<res_u; i++){
	    p_buffer->shorts_list[i].count = 3;
	    p_buffer->shorts_list[i].shorts = &ndx[j];
	    ndx[j++] = 0;	/* the pole position */
	    ndx[j++] = i+1;
	    ndx[j++] = i+2;
	}
	ndx[j-1] = 1;	/* adjust last index */

	cur_deg_v = delta_deg_v;
	cur_deg_u = 0.0;
	cos_cur_v = cos(cur_deg_v);
	f_const = y + radius*cos_cur_v;
	cur_rad = radius*sin(cur_deg_v);
	for( i = 0; i < res_u; i++, ptr++){
		n = (i+1) % res_u;
		ptr->x = x + cur_rad*p_build->cos_buffer_u[n];
		ptr->y = f_const;
		ptr->z = z - cur_rad*p_build->sin_buffer_u[n];
		cur_deg_u += delta_deg_u;
	}

	PEXSetOfFillAreaSets( dpy, p_state->cur_id, p_state->req_type,
		PEXShapeConvex, 0, 0, 0, PEXContourDisjoint, True, 
		p_state->color_type, res_u, facetList, res_u+1, vertList, 
		(res_u*3), (PEXSwitch *) NULL, p_buffer->cdata);

	/*
	 * move the top 'circle' generated vertices to every other location
	 * in the buffer for the triangle strip. This is needed to because
	 * of the 'down-up-down' nature of the tristrip.
	 * Start with the last point on the 'circle' of constant longitude.
	 */
	f_ptr = buf + res_u;
	ptr = buf + 2 * (res_u - 1);
	for( i = 0; i<res_u;i++, ptr -= 2, f_ptr--)
	    memcpy( ptr, f_ptr, sizeof(PEXCoord));

	cur_deg_v += delta_deg_v;
	f_ptr = buf;
	for( i = 1; i < res_v-1; i++){
	    ptr = buf + 1;
	    n = (i+1) % res_v;
	    cos_cur_v = p_build->cos_buffer_v[n];
	    sin_cur_v = p_build->sin_buffer_v[n];
	    cur_rad = radius*sin_cur_v;
	    f_const = y + radius*cos_cur_v;
	    cur_deg_u = 0;
	    for( j = 0; j < res_u; j++ ){
		n = (j+1) % res_u;
		ptr->x = x + cur_rad*p_build->cos_buffer_u[n];
		ptr->y = f_const;
		ptr->z = z - cur_rad*p_build->sin_buffer_u[n];
		cur_deg_u += delta_deg_u;
		ptr = ptr + 2;
	    }
	    /* Don't save duplicated last point */
	    /* ptr->x = f_ptr->x; ptr->y = f_ptr->y; ptr->z = f_ptr->z; */
	    cur_deg_v += delta_deg_v;
	    f_ptr = buf;
	    vertList.no_data = (PEXCoord *) buf;
	    ptr = ptr - 1; /* dupl first two verts at end of buffer */
	    memcpy( ptr, f_ptr, 2*sizeof(PEXCoord));
	    PEXTriangleStrip( dpy, p_state->cur_id, p_state->req_type, 0, 0, 
		p_state->color_type, facetList, (res_u+1)*2, vertList);
	    /* shift every 'new' vertex down one slot */
	    f_ptr = buf+1;
	    ptr = buf;
	    for( j = 0; j<res_u;j++, ptr += 2, f_ptr += 2 )
		memcpy( ptr, f_ptr, sizeof(PEXCoord));
	}

	/*
	 * move the bottom 'circle' generated vertices from every other
	 * location to every vertex adjacent to each other. The new 'circle'
	 * will begin at the second location in the buffer. The first
	 * location will be the bottom pole.
	 * Start with the second point on the 'circle' of constant longitude
	 * since the first point is at the correct location.
	 */
	f_ptr = buf + 3;
	ptr = buf + 2;
	for( i = 1; i<=res_u;i++, ptr++, f_ptr += 2 )
	    memcpy( ptr, f_ptr, sizeof(PEXCoord));

	buf->x = x; buf->y = y - radius; buf->z = z;

	/* do the indicies for the bottom section */

	for(i=0,j=0; i<res_u; i++){
	    p_buffer->shorts_list[i].count = 3;
	    p_buffer->shorts_list[i].shorts = &ndx[j];
	    ndx[j++] = 0;	/* the pole position */
	    ndx[j++] = i+2;
	    ndx[j++] = i+1;
	}
	ndx[j-2] = 1;	/* adjust last index */

	PEXSetOfFillAreaSets( dpy, p_state->cur_id, p_state->req_type,
		PEXShapeConvex, 0, 0, 0, PEXContourDisjoint, True, 
		p_state->color_type, res_u, facetList, res_u+1, vertList, 
		(res_u*3), (PEXSwitch *) NULL, p_buffer->cdata);
#ifdef KPC
	}
#endif
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pset_precision      +------------------------ *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the geometry building precision values.
 *
 *  USAGE:
 *	Pset_precision( ures, vres )
 *	int	    uprec	(in)	precision in 'u' direction
 *	int	    vprec	(in)	precision in 'v' direction
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pset_precision( uprec, vprec )
int	uprec;
int	vprec;
{
	p_state->attr->u_prec = uprec;
	p_state->attr->v_prec = vprec;
	PEXtk_build_cosine_buffer(uprec, vprec);
}

/*
 *
 *                       +---------------------------+
 * ----------------------+ PEXtk_build_cosine_buffer +------------------------
 *                       +---------------------------+
 *
 *  PURPOSE:
 *	Builds the cosine and sine buffer for the circle primitives.
 *
 *  USAGE:
 *	PEXtk_build_cosine_buffer( res_u, res_v )
 *	int	    res_u	(in)	The u-dir resolution
 *	int	    res_v	(in)	The v-dir resolution
 *
 *  RETURNS:
 *	0 if successful, 1 if error
 *
 *  DESCRIPTION:
 *	The sine and cosine buffer for the u-dir is based on a 360 rotation
 *	around the circumference.
 *	The sine and cosine buffer for the v-dir is based on a 180 rotation
 *	from the top pole to the bottom pole.
 */
 
/*ARGSUSED*/
int
PEXtk_build_cosine_buffer( res_u, res_v)
int	res_u;
int	res_v;
{
    int i;
    double cur_val, delta_rad;
    if( res_u != p_build->trig_buf_u_size ){
	if( p_build->cos_buffer_u )
	    free( (char *) p_build->cos_buffer_u);
	if(( p_build->cos_buffer_u = (double*)malloc( sizeof(double)*res_u)) 
		== NULL)
	    return 1;
	if( p_build->sin_buffer_u )
	    free( (char *) p_build->sin_buffer_u);
	if(( p_build->sin_buffer_u = (double*)malloc( sizeof(double)*res_u))
		== NULL)
	    return 1;
	p_build->trig_buf_u_size = res_u;
	delta_rad = (double) ((double) 2.0 * M_PI  / (double) res_u);
	cur_val = delta_rad;
	p_build->cos_buffer_u[0] = 1.0;
	p_build->sin_buffer_u[0] = 0.0;
	for( i=1; i<res_u; i++){
	    p_build->cos_buffer_u[i] = cos(cur_val);
	    p_build->sin_buffer_u[i] = sin(cur_val);
	    cur_val += delta_rad;
	}
	
    }
    if( res_v != p_build->trig_buf_v_size ){
	if( p_build->cos_buffer_v )
	    free( (char *) p_build->cos_buffer_v);
	if(( p_build->cos_buffer_v = (double*)malloc( sizeof(double)*res_v)) 
		== NULL)
	    return 1;
	if( p_build->sin_buffer_v )
	    free( (char *) p_build->sin_buffer_v);
	if(( p_build->sin_buffer_v = (double*)malloc( sizeof(double)*res_v))
		== NULL)
	    return 1;
	p_build->trig_buf_v_size = res_v;
	delta_rad = (double) ((double) M_PI  / (double) res_v);
	cur_val = delta_rad;
	p_build->cos_buffer_v[0] = 1.0;
	p_build->sin_buffer_v[0] = 0.0;
	for( i=1; i<res_v; i++){
	    p_build->cos_buffer_v[i] = cos(cur_val);
	    p_build->sin_buffer_v[i] = sin(cur_val);
	    cur_val += delta_rad;
	}
	
    }

    return 0;
}

int nvnorm, *pvnorm;
Fcoord3D *vl, *norm, *rnorm, *pnorm;

PEXtk_normalize(tn)
Fcoord3D *tn;
{
  float fac;

  fac = tn->x*tn->x + tn->y*tn->y + tn->z*tn->z;
  fac = (float) sqrt((double)fac);
  tn->x= tn->x * fac; tn->y= tn->y * fac; tn->z= tn->z * fac;
} 

/*
 *
 *                      +-----------------------------+
 * ---------------------+  PEXtk_calc_vertex_normals  +------------------------
 *                      +-----------------------------+
 *
 *  PURPOSE:
 *	Calculates the vertex normals.
 *
 *  USAGE:
 *      PEXtk_calc_vertex_normals(nverts,npolys,vl,fC,edgelist,vnorm)
 *	int	    nverts	(in)	number of vertices in vertex list
 *	int	    npolys	(in)	number of polygons in polygon list pl
 *	Fcoord3D    *vl		(in)	vertex list
 *	short	    *fC		(in)	facet count list
 *	short	    *edgelist	(in)	vertex indices/edge connectivity list
 *	Fcoord3D    **vnorm	(out)	vertex normal list
 *
 *  NOTE:
 *	The vertex normals are returned in allocated memory at *vnorm. The
 *	calling routine must free this when it is done.
 *	The method used is a brute force method:
 *		foreach vertex
 *		    foreach polygon
 *			if vertex is in this polygon
 *			    add polygon normal x,y,z to vertex normal x,y,z
 *
 *  RETURNS:
 *	0 if error, non-zero if successful (number of vertex normals)
 */
 
/*ARGSUSED*/
int
PEXtk_calc_vertex_normals( nverts, npolys, vl, fC, edgelist, vnorm)
int	nverts;		/* number of vertices in vertex list */
int	npolys;		/* number of polygons in polygon list pl */
Fcoord3D *vl;		/* vertex list */
short	*fC;		/* facet count list */
short	*edgelist;	/* vertex indices/edge connectivity list */
Fcoord3D **vnorm;	/* vertex normal list */
{ 
  int nv, pn, vn, i, avn;
  float avx,avy,avz;
  Fcoord3D tn;
  short *pl;

  if( PEXtk_calc_polygon_normals(nverts,npolys,vl,fC,edgelist))
	return 0;

  /* get enough memory for vertex normals */
  if(( *vnorm = (Fcoord3D *)malloc(nverts*sizeof(Fcoord3D))) == NULL)
	return 0;
  nvnorm = 0;

    rnorm = *vnorm;
  for(vn=0;vn<nverts;vn++){	/* for every vertex */
    pl = edgelist; avn = 0; avx = avy = avz = 0;
    rnorm->x = rnorm->y = rnorm->z = 0.0;

    for(pn=0;pn<npolys;pn++){   /* search every poly for use of this vertex */
      nv = fC[pn];		/* number of vertices in this poly */
      for(i=0;i<nv;i++)		/* look at every vertex in this polygon */
	if(*pl++ == vn){ 	/* avg. poly's normal if in this poly */
	  avn++;
	  avx += pnorm[pn].x; avy += pnorm[pn].y;
  	  avz += pnorm[pn].z;
	  pl  += nv-i-1;
	  i    = nv;		/* go to next polygon */
	}
    }
    if(avn){
	tn.x= avx/avn; tn.y= avy/avn; tn.z= avz/avn;
	PEXtk_normalize(&tn);		/* re-normalize the vector */
	rnorm->x= tn.x; rnorm->y= tn.y; rnorm->z= tn.z;
	nvnorm++;
    }
    rnorm++;
  }

  free((char *) pnorm);
  return(nvnorm);
}

/*
 *
 *                      +------------------------------+
 * ---------------------+  PEXtk_calc_polygon_normals  +-----------------------
 *                      +------------------------------+
 *
 *  PURPOSE:
 *	Calculate the polygon normals as accurately as possible
 *
 *  USAGE:
 *      PEXtk_calc_polygon_normals(nverts,npolys,vl,fC,edgelist)
 *	int	    nverts	(in)	number of vertices in vertex list
 *	int	    npolys	(in)	number of polygons in polygon list pl
 *	Fcoord3D    *vl		(in)	vertex list
 *	short	    *fC		(in)	facet count list
 *	short	    *edgelist	(in)	vertex indices/edge connectivity list
 *
 *  NOTE:
 *	The polygon normals are returned in allocated memory *pnorm. The
 *	calling routine must free this when it is done.
 *	If polygon does not have a normal, then the polygon normal is
 *	set to 0.0.
 *
 *  RETURNS:
 *	0 if successful, non-zero if error
 */
 
/*ARGSUSED*/
int
PEXtk_calc_polygon_normals(nverts,npolys,vl,fC,edgelist)
int	nverts;		/* number of vertices in vertex list */
int	npolys;		/* number of polygons in polygon list pl */
Fcoord3D *vl;		/* vertex list */
short	*fC;		/* facet count list */
short	*edgelist;	/* vertex indices/edge connectivity list */
{
    double dx1, dy1, dz1, dx2, dy2, dz2, xn, yn, zn;
    double x[3], y[3], z[3];
    double fac;
    int i, j, nv;
    short *p;

  /* get enough memory for polygon normals */
  if(( pnorm = (Fcoord3D *)malloc(npolys* sizeof(Fcoord3D))) == NULL)
	return 1;

  /*
   * For each polygon,
   *	get number of verts in edge list from facetcount list
   *	traverse the edge list to:
   *	    get first 3 vertices to calc polygon normal
   *	    after first 3 vertices, continue to traverse to get to next poly
   */
  p = edgelist;
    for(i=0;i<npolys;i++){ /* for all polygons */

	nv = fC[i];		/* number of vertices */
	for(j=0;j<nv;j++){	/* for each vertex */
	    if(j<3){		/* get first 3 points */
		x[j]= vl[*p].x; y[j]= vl[*p].y; z[j]= vl[*p].z;
		p++;
	    } else p++;
	} /* end for */

	/* get cross product */
	dx1= x[0] - x[1]; dy1= y[0] - y[1]; dz1= z[0] - z[1];
	dx2= x[1] - x[2]; dy2= y[1] - y[2]; dz2= z[1] - z[2];
	xn= (dy1 * dz2) - (dz1 * dy2); yn= (dz1 * dx2) - (dx1 * dz2);
	zn= (dx1 * dy2) - (dy1 * dx2);

	/* and normalize */
	fac = sqrt( xn*xn + yn*yn + zn*zn);
	if( fac != 0.0 ){
	    fac = 1.0 / fac;
	    pnorm[i].x = (float) (xn * fac);
	    pnorm[i].y = (float) (yn * fac);
	    pnorm[i].z = (float) (zn * fac);
	} else {
	    pnorm[i].x = (float) (0.0);
	    pnorm[i].y = (float) (0.0);
	    pnorm[i].z = (float) (0.0);
	}
    } /* end for all polygons */
    return 0;
}
