/*
 ******************************************************************************
 ******************************************************************************
 *****                          clr_test.c                                *****
 *****									  *****
 *****			PEXtk Color test Program  			  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1991, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA      94086  			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************** */

/*
	Program:   clr_test.c

	Synopsis   This program is a simple example of using PEXtk
		   with a custom colormap. 

	Usage:	   clr_test [options] 

	    where:
		options are:
		    -d 			turns on double buffering
		    -i num		sets iteration count
		    -c 			uses colors per vertex
		    -n 			uses pre-calculated polygon normals
		    -w size		sets the size of the window

*/

#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <PEXtk.h>
#include <math.h>

    float polyData[]={0.0,0.0,1.0,0.0,0.0,1.0};
    unsigned short polyCounts[]={4,4};
    unsigned short edges[] = {0,1,2,3,3,2,4,5};
    float vertexData[] = {-5.0,5.0,-4.0,
                          -5.0,-5.0,-4.0,
                           0.0,-5.0,-4.0,
                           0.0,5.0,-4.0,
                           5.0,-5.0,-4.0,
                           5.0,5.0,-4.0};

    /* vertex color data, red on left, green in center, blue on right */
    float vertexClrData[] = {-5.0,5.0,-4.0,1.0,0.0,0.0,
                          -5.0,-5.0,-4.0,1.0,0.0,0.0,
                           0.0,-5.0,-4.0,0.0,1.0,0.0,
                           0.0,5.0,-4.0,0.0,1.0,0.0,
                           5.0,-5.0,-4.0,0.0,0.0,1.0,
                           5.0,5.0,-4.0,0.0,0.0,1.0};

    int			 win_width, win_height;
    static int 		 lkup_rgb[256][3];
    static unsigned long lkup_rgb_pix[256];
	int num_red, num_green, num_blue;
	int mult_red, mult_green, mult_blue;
	long	Get_Pseudo_Color();
	long	Get_Index_Color();

main(argc,argv)
int	argc;
char	*argv[];
{
	Display *dpy; 
	Window window;
	Visual *win_visual;
	int screen_num;
	Colormap cmap;
	int i,iter=40,dbl_buf, clr_flag=0,norm_flag=0;
	PCOLOR	col;

	win_width=500; win_height=500;

	dbl_buf = 0;
	if( getenv("DBL_BUF") )
	    dbl_buf = 1;

	for(i=1;i<argc;i++){
	    if( strcmp(argv[i],"-i") == 0 ){
		i++;
		iter = atoi(argv[i]);
	    }
	    if( strcmp(argv[i],"-d") == 0 ){
		dbl_buf = 1;
	    }
	    if( strcmp(argv[i],"-c") == 0 ){
		clr_flag = 1;
	    }
	    if( strcmp(argv[i],"-n") == 0 ){
		norm_flag = GANORMAL;
	    }
	    if( strcmp(argv[i],"-w") == 0 ){
		i++;
		win_width = atoi(argv[i]);
		win_height = win_width;
	    }
	}

        if ((dpy = XOpenDisplay ("")) == NULL) {
            printf(" Error opening display\n");
	    return 1;
        }

	if(Pinit(dpy)){
	    printf("ERROR... initializing PEX server\n");
	    exit(1);
	}

	if( (window = Pcreate_window(dpy, win_width, win_height)) == NULL){
	    printf("ERROR... creating window\n");
	    exit(1);
	}

	screen_num = DefaultScreen(dpy);
	win_visual = DefaultVisual( dpy, screen_num);

	Get_Visual_Type( dpy, &win_visual, PseudoColor);

	cmap = XCreateColormap( dpy, window,
		win_visual, AllocNone);
        printf("Visual class %d, true %d, id %x\n",win_visual->class,TrueColor,
			win_visual->visualid);

	num_red = num_green = num_blue = 6;
	mult_red = 36; mult_green = 6; mult_blue = 1;

	Pset_function( PEXtk_FUNC_PIXEL_RGB, Get_Pseudo_Color);
	Pset_function( PEXtk_FUNC_PIXEL_INDEX, Get_Index_Color);

	Setup_Colormap( dpy, window, cmap, 0,
		num_red, num_green, num_blue, 2.3 );

	Pset_colormap( cmap);
	Pset_drawable(window);
	if( cmap == NULL )
	    printf("ERROR... XCreateColormap\n");
	else {
	    printf("CreateColormap %x, default %x\n",cmap,
		DefaultColormap(dpy,screen_num));
	    XSetWindowColormap( dpy, window,
				cmap);
	    i = 0; /* basePixel */
            Pset_color_equation( &cmap, &i, 6, 6, 6, 6*6, 6, 1, 2.3);
	}

	if( dbl_buf )
	    Pbuffer_mode(PEXtk_DOUBLE_BUFFER);
	else
	    Pbuffer_mode(PEXtk_SINGLE_BUFFER);

	col.color.rgb.red=1.0;col.color.rgb.green=col.color.rgb.blue=.4;

	Pzbuffer(1);
	Pbackface(PEXtk_CULL_NONE);
	Pviewport( 0, win_width-1, 0, win_height-1);
	Ppersp_view( 0, 90.0, 1.0, 1.0, 1000.0);
	Pclear_rgb( 0.0, 0.0, 0.0);

	if( clr_flag && DefaultDepth(dpy,screen_num) != 8 )
	    Pshade_mode( PEXtk_SHADE_NORMALS | PEXtk_SHADE_COLORS );
	else
	    Pshade_mode( PEXtk_SHADE_NORMALS );

	Create_str(clr_flag);
        Ptranslate( 0., 0., -50., 1);
	for(i=0;i<iter;i++)
	{
	    Pclear_rgb( 0.0, 0.0, 0.0);
	    Pline_color(&col);
	    Pcolor(&col);
	    Pexec_struct(1);
	    Protate_y( 4.0, 1);
	    if( dbl_buf ){
		Pswap_buffers();
		Pflush(0);
	    } else
		Pflush(1);
	}
	Pflush(1);
	printf("Press return\n");scanf("%c",&iter);
	Pexit();
	XDestroyWindow( dpy, window);
	XCloseDisplay(dpy);
	exit(0);
}

Create_str(clr_flag)
int clr_flag;
{
    Pcreate_struct(1);
    if( clr_flag )
	Ppoly_index( GANORMAL, GACOLOR, 2, 6, 0, PEXtk_COLOR_RGB, polyCounts,
                polyData, edges, vertexClrData);
    else
	Ppoly_index( GANORMAL, 0, 2, 6, 0, PEXtk_COLOR_RGB, polyCounts,
                polyData, edges, vertexData);
#ifdef NEW
    if( clr_flag )
	Ppoly_index( GANORMAL, GACOLOR, 2, 6, 0, PEXtk_COLOR_RGB, polyCounts,
                polyData, poly_lists, vertexClrData);
    else
	Ppoly_index( GANORMAL, 0, 2, 6, 0, PEXtk_COLOR_RGB, polyCounts,
                polyData, poly_lists, vertexData);
#endif
    Pclose_struct();
}

Get_Visual_Type(dpy,vis,type)
Display *dpy;
Visual **vis;
int	type;
{
   XVisualInfo vTemplate, *visualList;
   int i, n_visuals, class, screen;
   Visual *win_vis;

   screen = DefaultScreen (dpy);

    vTemplate.screen = screen;
    /* vTemplate.depth = 8; */

    visualList = XGetVisualInfo( dpy, VisualScreenMask,
        &vTemplate, &n_visuals);

    if( n_visuals == 0 ){
	*vis = DefaultVisual( dpy, screen);
	return;
    }

    for(i=0;i<n_visuals;i++){
#if defined(__cplusplus) || defined(c_plusplus)
        if( visualList[i].c_class == type)
#else
        printf("class %d, look for %d\n", visualList[i].class,type);
        if( visualList[i].class == type)
#endif
	{
	    *vis = visualList[i].visual;
	    return;
	}
    }
    *vis = visualList[0].visual;
    return;
}

int 
Setup_Colormap(dpy,window,cmap,basePixel,num1,num2,num3,gamma_value)
Display		*dpy;
Window		window;
Colormap	cmap;
int		basePixel;
int		num1;
int		num2;
int		num3;
float		gamma_value;
{

	Window		root;		/* root window if display */
	Atom		cmapAtom=0;
	unsigned long	pixels[256];
	unsigned long	planeMasks[24];
	unsigned short	Gamma[256];
	int		r, g, b;
	int		i, status;
	double		rinc, ginc, binc;
	int		two_to_16th = 1 << 16;
	double		igamma, value, num, inc;
	int		maxIntensity;
	XColor		pix_color[256];
	int		breakPixel;	/* for finding contiguous cells */
	int		colorCount;
	int		baseCount=0;


	colorCount = (num1) * (num2) * (num3);
	baseCount = basePixel;
	maxIntensity = CellsOfScreen(DefaultScreenOfDisplay(dpy)) - 1;

	    status = XAllocColorCells(dpy, cmap, 1, planeMasks, 0, 
			&pixels[baseCount], colorCount);
	if( status == 0 )printf("ERROR... status = 0\n");

	for (i=0,breakPixel=0;((i<colorCount-1) && (!breakPixel));i++) {
		if (pixels[i] != pixels[i+1]-1)
		   breakPixel = i+1;
	}
		
	for (i = baseCount; i < colorCount; i++) {
		pixels[i] = pixels[i+breakPixel];
		pix_color[i].pixel = pixels[i];
	}
/* printf("Base %d\n", pixels[i+breakPixel]); */

	rinc = maxIntensity / (double)( num1 );
	ginc = maxIntensity / (double)( num2 );
	binc = maxIntensity / (double)( num3 );

	igamma = 1.0/gamma_value;
	inc = (double) two_to_16th / (double) maxIntensity;
	num = inc / 2.0;
	for (i = 0; i <= maxIntensity; i++) {
		value = num / (double) (two_to_16th - 1);
		Gamma[i] =
			(unsigned short) (pow(value, igamma) * two_to_16th);
		num += inc;
	}
	i=0;
	for (r = 0; r < (int) num1; r++)
	    for (g = 0; g < (int) num2; g++)
		for (b = 0; b < (int) num3; b++) {
		    pix_color[i+baseCount].red   = 
			Gamma[(int)((double)r * rinc)];
		    pix_color[i+baseCount].green = 
			Gamma[(int)((double)g * ginc)];
		    pix_color[i+baseCount].blue  = 
			Gamma[(int)((double)b * binc)];
		    pix_color[i+baseCount].flags = DoRed | DoGreen | DoBlue;
		    lkup_rgb[i][0]= r*maxIntensity/(num1-1);
		    lkup_rgb[i][1]= g*maxIntensity/(num2-1);
		    lkup_rgb[i][2]= b*maxIntensity/(num3-1);
		    i++;
		}

	/* for(i=0;i<baseCount+colorCount;i++) */
	for(i=0;i<baseCount;i++)
	{
	    pix_color[i].pixel = i;
	    XQueryColor( dpy, cmap, &pix_color[i]);
	}
	for(i=0;i<colorCount-baseCount;i++){
	    lkup_rgb_pix[i] = pix_color[i+baseCount].pixel;
	}
#ifdef PRINT_CMAP
	for(i=0;i<colorCount;i++){
	    printf(" pix %d i %d rgb %d %d %d : def %d, %d %d %d\n",
		lkup_rgb_pix[i], i,lkup_rgb[i][0],lkup_rgb[i][1],lkup_rgb[i][2],
		pix_color[i].pixel,pix_color[i].red,pix_color[i].green,
		pix_color[i].blue);
	}
#endif
	XStoreColors(dpy, cmap,pix_color,colorCount);
}

long
Get_Index_Color( index )
{
   return (long) lkup_rgb_pix[index];
}

long
Get_Pseudo_Color( r, g, b )
float r,g,b;
{
    int index;
    long pix;
        index = (int)
          ((int)((int)((r)*(num_red-1))    * mult_red) +
          ((int)((int)((g)*(num_green-1))) * mult_green) +
          ((int)((int)((b)*(num_blue-1)))  * mult_blue));

   return (long) lkup_rgb_pix[index];
}
